# Standard library imports
from typing import Any

def dumps(
    obj: Any,
    skipkeys: bool = ...,
    ensure_ascii: bool = ...,
    check_circular: bool = ...,
    allow_nan: bool = ...,
    cls: Any | None = ...,
    indent: int | None = ...,
    separators: tuple[str, str] | None = ...,
    default: Any | None = ...,
    sort_keys: bool = ...,
    quote_keys: bool = ...,
    trailing_commas: bool = ...,
    allow_duplicate_keys: bool = ...,
) -> str: ...
